/*
 * Copyright © Original Authors
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional copyright and licensing notices may apply for content that was
 * included from other projects. For more information, see ATTRIBUTION.md.
 */

package grondag.xm.terrain;

import static grondag.xm.api.modelstate.ModelStateFlags.NONE;

import java.util.function.Consumer;
import net.minecraft.class_2350;
import net.minecraft.class_2960;
import org.jetbrains.annotations.ApiStatus.Internal;
import grondag.xm.Xm;
import grondag.xm.api.mesh.WritableMesh;
import grondag.xm.api.mesh.XmMesh;
import grondag.xm.api.mesh.XmMeshes;
import grondag.xm.api.mesh.polygon.MutablePolygon;
import grondag.xm.api.mesh.polygon.Polygon;
import grondag.xm.api.paint.SurfaceTopology;
import grondag.xm.api.primitive.surface.XmSurface;
import grondag.xm.api.primitive.surface.XmSurfaceList;
import grondag.xm.api.terrain.TerrainModelState;

@Internal
public class TerrainCubePrimitive extends AbstractTerrainPrimitive {
	public static final XmSurfaceList SURFACES = XmSurfaceList.builder().add("all", SurfaceTopology.CUBIC, XmSurface.FLAG_ALLOW_BORDERS).build();

	public static final XmSurface SURFACE_ALL = SURFACES.get(0);

	public static final TerrainCubePrimitive INSTANCE = new TerrainCubePrimitive(Xm.id("terrain_cube"));

	/** Never changes so may as well save it. */
	private final XmMesh cachedQuads;

	protected TerrainCubePrimitive(class_2960 id) {
		super(id, NONE, TerrainModelStateImpl.FACTORY, s -> SURFACES);
		cachedQuads = getCubeQuads();
	}

	@Override
	public void emitQuads(TerrainModelState modelState, Consumer<Polygon> target) {
		cachedQuads.forEach(target);
	}

	private XmMesh getCubeQuads() {
		final WritableMesh stream = XmMeshes.claimWritable();
		appendCubeFace(stream, class_2350.field_11033);
		appendCubeFace(stream, class_2350.field_11036);
		appendCubeFace(stream, class_2350.field_11034);
		appendCubeFace(stream, class_2350.field_11039);
		appendCubeFace(stream, class_2350.field_11043);
		appendCubeFace(stream, class_2350.field_11035);

		final XmMesh result = stream.releaseToReader();

		result.reader().origin();
		assert result.reader().vertexCount() == 4;

		return result;
	}

	private void appendCubeFace(WritableMesh stream, class_2350 side) {
		final MutablePolygon q = stream.writer();

		q.lockUV(0, true);
		q.surface(SURFACE_ALL);
		q.nominalFace(side);

		switch (side) {
			case field_11036:
				q.vertex(0, 0, 1, 0, 0, 0, 0xFFFFFFFF);
				q.vertex(1, 0, 1, 1, 0, 1, 0xFFFFFFFF);
				q.vertex(2, 1, 1, 1, 1, 1, 0xFFFFFFFF);
				q.vertex(3, 1, 1, 0, 1, 0, 0xFFFFFFFF);
				break;

			case field_11033:
				q.vertex(0, 0, 0, 1, 1, 1, 0xFFFFFFFF);
				q.vertex(1, 0, 0, 0, 1, 0, 0xFFFFFFFF);
				q.vertex(2, 1, 0, 0, 0, 0, 0xFFFFFFFF);
				q.vertex(3, 1, 0, 1, 0, 1, 0xFFFFFFFF);
				break;

			case field_11039:
				q.vertex(0, 0, 1, 0, 0, 0, 0xFFFFFFFF);
				q.vertex(1, 0, 0, 0, 0, 1, 0xFFFFFFFF);
				q.vertex(2, 0, 0, 1, 1, 1, 0xFFFFFFFF);
				q.vertex(3, 0, 1, 1, 1, 0, 0xFFFFFFFF);
				break;

			case field_11034:
				q.vertex(0, 1, 1, 1, 0, 0, 0xFFFFFFFF);
				q.vertex(1, 1, 0, 1, 0, 1, 0xFFFFFFFF);
				q.vertex(2, 1, 0, 0, 1, 1, 0xFFFFFFFF);
				q.vertex(3, 1, 1, 0, 1, 0, 0xFFFFFFFF);
				break;

			case field_11043:
				q.vertex(0, 1, 1, 0, 0, 0, 0xFFFFFFFF);
				q.vertex(1, 1, 0, 0, 0, 1, 0xFFFFFFFF);
				q.vertex(2, 0, 0, 0, 1, 1, 0xFFFFFFFF);
				q.vertex(3, 0, 1, 0, 1, 0, 0xFFFFFFFF);
				break;

			case field_11035:
				q.vertex(0, 0, 1, 1, 0, 0, 0xFFFFFFFF);
				q.vertex(1, 0, 0, 1, 0, 1, 0xFFFFFFFF);
				q.vertex(2, 1, 0, 1, 1, 1, 0xFFFFFFFF);
				q.vertex(3, 1, 1, 1, 1, 0, 0xFFFFFFFF);
				break;
		}

		q.append();
	}

	@Override
	public TerrainModelState.Mutable geometricState(TerrainModelState fromState) {
		return defaultState().mutableCopy();
	}

	@Override
	public boolean doesShapeMatch(TerrainModelState from, TerrainModelState to) {
		return from.primitive() == to.primitive();
	}
}
