/*
 * Copyright © Original Authors
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional copyright and licensing notices may apply for content that was
 * included from other projects. For more information, see ATTRIBUTION.md.
 */

package grondag.xm.terrain;

import static it.unimi.dsi.fastutil.HashCommon.arraySize;

import java.util.function.Supplier;
import net.minecraft.class_1922;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_3610;
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap;
import org.jetbrains.annotations.ApiStatus.Internal;
import grondag.fermion.position.PackedBlockPos;

// TODO: add caching for flow height - do with as part of SuperBlockState
// TODO: reinstate usage or remove

/**
 * Caches expensive world state lookups until next prepare() call. For
 * server-side use on server thread only. Doesn't try to track state changes
 * while not in active use, and all state changes must be single threaded and
 * occur through this instance.
 */
@Internal
public class TerrainWorldAdapter implements class_1922 {
	protected class_1937 world;

	@SuppressWarnings("serial")
	public static class FastMap<V> extends Long2ObjectOpenHashMap<V> {
		/**
		 * DOES NOT SUPPORT ZERO-VALUED KEYS.
		 * Only computes key 1x and scans arrays 1x for a modest savings. Here because
		 * block updates are the on-tick performance bottleneck for lava sim.
		 */
		private V computeFast(final long k, final Supplier<V> v) {
			final long[] key = this.key;
			int pos = (int) it.unimi.dsi.fastutil.HashCommon.mix(k) & mask;
			long curr = key[pos];

			// The starting point.
			if (curr != 0) {
				if (curr == k) {
					return value[pos];
				}

				while (!((curr = key[pos = (pos + 1) & mask]) == (0)))
					if (curr == k) {
						return value[pos];
					}
			}

			final V result = v.get();
			key[pos] = k;
			value[pos] = result;

			if (size++ >= maxFill) {
				rehash(arraySize(size + 1, f));
			}

			return result;
		}
	}

	protected FastMap<class_2680> blockStates = new FastMap<>();
	protected FastMap<TerrainState> terrainStates = new FastMap<>();

	public TerrainWorldAdapter() { }

	public TerrainWorldAdapter(class_1937 world) {
		prepare(world);
	}

	public void prepare(class_1937 world) {
		this.world = world;
		blockStates.clear();
		terrainStates.clear();
	}

	public class_1937 wrapped() {
		return world;
	}

	@Override
	public class_2680 method_8320(final class_2338 pos) {
		final long packedBlockPos = PackedBlockPos.pack(pos);
		return blockStates.computeFast(packedBlockPos, () -> world.method_8320(pos));
	}

	private final class_2338.class_2339 getBlockPos = new class_2338.class_2339();

	public class_2680 getBlockState(long packedBlockPos) {
		return blockStates.computeFast(packedBlockPos, () -> {
			PackedBlockPos.unpackTo(packedBlockPos, getBlockPos);
			return world.method_8320(getBlockPos);
		});
	}

	private final class_2338.class_2339 getTerrainPos = new class_2338.class_2339();

	public TerrainState terrainState(class_2680 state, long packedBlockPos) {
		return terrainStates.computeFast(packedBlockPos, () -> {
			PackedBlockPos.unpackTo(packedBlockPos, getTerrainPos);
			return TerrainState.terrainState(this, state, getTerrainPos);
		});
	}

	public TerrainState terrainState(class_2680 state, class_2338 pos) {
		return terrainStates.computeFast(PackedBlockPos.pack(pos), () -> {
			return TerrainState.terrainState(this, state, pos);
		});
	}

	public TerrainState terrainState(long packedBlockPos) {
		return terrainState(getBlockState(packedBlockPos), packedBlockPos);
	}

	/**
	 * Note this doesn't invalidate terrain state cache. Need to do that directly
	 * before using anything that needs it if changing terrain surface.
	 */
	public void setBlockState(long packedBlockPos, class_2680 newState) {
		setBlockState(packedBlockPos, newState, true);
	}

	/**
	 * Use when you want to control
	 * {@link #onBlockStateChange(long, class_2680, class_2680)} call back.
	 */
	protected void setBlockState(long packedBlockPos, class_2680 newState, boolean callback) {
		final class_2680 oldState = getBlockState(packedBlockPos);

		if (newState == oldState) {
			return;
		}

		blockStates.put(packedBlockPos, newState);
		applyBlockState(packedBlockPos, oldState, newState);

		if (callback) {
			onBlockStateChange(packedBlockPos, oldState, newState);
		}
	}

	/**
	 * Handles application of block state to world. Override for deferred updates or
	 * use cases where world should not be affected directly.
	 */
	protected void applyBlockState(long packedBlockPos, class_2680 oldState, class_2680 newState) {
		world.method_8501(PackedBlockPos.unpack(packedBlockPos), newState);
	}

	/**
	 * Called for all block state changes, even if not a terrain block.
	 */
	protected void onBlockStateChange(long packedBlockPos, class_2680 oldBlockState, class_2680 newBlockState) {
		// NOOP
	}

	public void setBlockState(class_2338 blockPos, class_2680 newState) {
		this.setBlockState(PackedBlockPos.pack(blockPos), newState);
	}

	@Override
	public class_2586 method_8321(class_2338 pos) {
		return world.method_8321(pos);
	}

	@Override
	public class_3610 method_8316(class_2338 pos) {
		return world.method_8316(pos);
	}

	@Override
	public int method_31605() {
		return world.method_31605();
	}

	@Override
	public int method_31607() {
		return world.method_31607();
	}
}
