/*
 * Copyright © Original Authors
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional copyright and licensing notices may apply for content that was
 * included from other projects. For more information, see ATTRIBUTION.md.
 */

package grondag.xm.virtual;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2596;
import net.minecraft.class_3222;
import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import org.jetbrains.annotations.ApiStatus.Internal;
import org.jetbrains.annotations.Nullable;
import io.vram.sc.unordered.SimpleUnorderedArrayList;

import grondag.fermion.position.IntegerBox;
import grondag.xm.Xm;
import grondag.xm.XmConfig;
import grondag.xm.network.S2C_ExcavationRenderUpdate;

/**
 * Class exists on server but render methods do not. Server instantiates (and
 * generates IDs) and transmits to clients.
 */
@Internal
public class ExcavationRenderEntry {
	private static int nextID = 0;

	public final int id;

	public final class_1937 world;

	public final ExcavationRenderTask task;

	@Nullable
	private IntegerBox aabb;

	private boolean isFirstComputeDone = false;

	/** Non-null if should render individual renderPositions instead of AABB. */
	@Nullable
	private class_2338[] renderPositions = null;

	private final Int2ObjectMap<AtomicInteger> xCounts = new Int2ObjectOpenHashMap<>();
	private final Int2ObjectMap<AtomicInteger> yCounts = new Int2ObjectOpenHashMap<>();
	private final Int2ObjectMap<AtomicInteger> zCounts = new Int2ObjectOpenHashMap<>();

	private final Set<class_2338> positions = Collections.synchronizedSet(new HashSet<class_2338>());

	private boolean isValid = true;

	/**
	 * If true, has changed after the start of the last computation. Cleared at
	 * start of computation run.
	 *
	 * <p>If dirty when computation completes, computation will resubmit self to queue
	 * for recomputation.

	 * If becomes dirty while computation not in progress, {@link #setDirty()} will
	 * submit for computation.
	 */
	private final AtomicBoolean isDirty = new AtomicBoolean(true);

	/**
	 * Players who could be viewing this excavation and should received client-side
	 * updates.
	 */
	private final SimpleUnorderedArrayList<class_3222> listeners = new SimpleUnorderedArrayList<>();

	private void addPos(class_2338 pos) {
		positions.add(pos);

		synchronized (xCounts) {
			AtomicInteger xCounter = xCounts.get(pos.method_10263());

			if (xCounter == null) {
				xCounter = new AtomicInteger(1);
				xCounts.put(pos.method_10263(), xCounter);
			} else {
				xCounter.incrementAndGet();
			}
		}

		synchronized (yCounts) {
			AtomicInteger yCounter = yCounts.get(pos.method_10264());

			if (yCounter == null) {
				yCounter = new AtomicInteger(1);
				yCounts.put(pos.method_10264(), yCounter);
			} else {
				yCounter.incrementAndGet();
			}
		}

		synchronized (zCounts) {
			AtomicInteger zCounter = zCounts.get(pos.method_10260());

			if (zCounter == null) {
				zCounter = new AtomicInteger(1);
				zCounts.put(pos.method_10260(), zCounter);
			} else {
				zCounter.incrementAndGet();
			}
		}
	}

	/**
	 * Returns true if any dimension had a count drop to zero.
	 */
	private boolean removePos(class_2338 pos) {
		positions.remove(pos);
		boolean gotZero = xCounts.get(pos.method_10263()).decrementAndGet() == 0;
		gotZero = yCounts.get(pos.method_10264()).decrementAndGet() == 0 || gotZero;
		gotZero = zCounts.get(pos.method_10260()).decrementAndGet() == 0 || gotZero;
		return gotZero;
	}

	/**
	 * For server side.
	 */
	public ExcavationRenderEntry(ExcavationRenderTask task) {
		id = nextID++;
		world = task.world();
		this.task = task;

		if (XmConfig.logExcavationRenderTracking) {
			Xm.LOG.info("id = %d new Entry constructor", id);
		}

		task.addCompletionListener(this::onPositionComplete);

		task.forEachPosition(this::addPos);

		if (ExcavationRenderEntry.this.positions.size() == 0) {
			if (XmConfig.logExcavationRenderTracking) {
				Xm.LOG.info("id = %d new Entry constructor - invalid", id);
			}

			isValid = false;
		} else {
			if (XmConfig.logExcavationRenderTracking) {
				Xm.LOG.info("id = %d new Entry constructor - launching compute", id);
			}

			ExcavationRenderEntry.this.compute();
		}
	}

	public void onPositionComplete(class_2338 pos) {
		final boolean needsCompute = removePos(pos);
		isValid = isValid && positions.size() > 0;

		if (isValid) {
			if (needsCompute) {
				setDirty();
			}
		} else {
			ExcavationRenderTracker.INSTANCE.remove(this);
		}
	}

	private void setDirty() {
		isDirty.compareAndSet(false, true);
		compute();
	}

	/**
	 * If false, can't send packets with this. Implies compute in progress or to be
	 * scheduled.
	 */
	public boolean isFirstComputeDone() {
		return isFirstComputeDone;
	}

	public void compute() {
		if (XmConfig.logExcavationRenderTracking) {
			Xm.LOG.info("id = %d Compute running.", id);
		}

		isDirty.set(false);

		final int count = positions.size();

		if (count == 0) {
			if (XmConfig.logExcavationRenderTracking) {
				Xm.LOG.info("id = %d Compute existing due to empty positions.", id);
			}

			updateListeners();
			ExcavationRenderTracker.INSTANCE.remove(this);
			return;
		}

		int minX = Integer.MAX_VALUE;
		int maxX = Integer.MIN_VALUE;

		for (final Int2ObjectMap.Entry<AtomicInteger> x : xCounts.int2ObjectEntrySet()) {
			if (x.getValue().get() > 0) {
				minX = Math.min(minX, x.getIntKey());
				maxX = Math.max(maxX, x.getIntKey());
			}
		}

		int minY = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;

		for (final Int2ObjectMap.Entry<AtomicInteger> y : yCounts.int2ObjectEntrySet()) {
			if (y.getValue().get() > 0) {
				minY = Math.min(minY, y.getIntKey());
				maxY = Math.max(maxY, y.getIntKey());
			}
		}

		int minZ = Integer.MAX_VALUE;
		int maxZ = Integer.MIN_VALUE;

		for (final Int2ObjectMap.Entry<AtomicInteger> z : zCounts.int2ObjectEntrySet()) {
			if (z.getValue().get() > 0) {
				minZ = Math.min(minZ, z.getIntKey());
				maxZ = Math.max(maxZ, z.getIntKey());
			}
		}

		final IntegerBox newBox = new IntegerBox(minX, minY, minZ, maxX + 1, maxY + 1, maxZ + 1);

		// always send start time computed
		boolean needsListenerUpdate = !isFirstComputeDone;
		isFirstComputeDone = true;

		if (!newBox.equals(aabb)) {
			aabb = newBox;
			needsListenerUpdate = true;
		}

		if (count <= 16 && (renderPositions == null || renderPositions.length != count)) {
			synchronized (positions) {
				class_2338[] newPositions = new class_2338[positions.size()];
				newPositions = positions.toArray(newPositions);
				renderPositions = newPositions;
			}

			needsListenerUpdate = true;

			if (XmConfig.logExcavationRenderTracking) {
				Xm.LOG.info("id %d Computed render position length = %d", id, renderPositions == null ? 0 : renderPositions.length);
			}
		}

		if (XmConfig.logExcavationRenderTracking) {
			Xm.LOG.info("id = %d Compute done, updateListeners=%s, isDirty=%s", id, Boolean.toString(needsListenerUpdate),
				Boolean.toString(isDirty.get()));
		}

		if (needsListenerUpdate) {
			updateListeners();
		}

		if (isDirty.get() && count > 0) {
			compute();
		}
	}

	/**
	 * Checked by excavation tracker on creation and will not add if false.
	 * {@link #onTaskComplete(AbstractTask)} also uses as signal to remove this
	 * instance from tracker.
	 */
	public boolean isValid() {
		return isValid;
	}

	public IntegerBox aabb() {
		return aabb;
	}

	public void addListener(class_3222 listener, boolean sendPacketIfNew) {
		if (XmConfig.logExcavationRenderTracking) {
			Xm.LOG.info("id=%d addListenger sendIfNew=%s, isValue=%s, isFirstComputeDone=%s", id, Boolean.toString(sendPacketIfNew),
				Boolean.toString(isValid), Boolean.toString(isFirstComputeDone));
		}

		synchronized (listeners) {
			if (listeners.addIfNotPresent(listener) && sendPacketIfNew && isValid && isFirstComputeDone) {
				if (XmConfig.logExcavationRenderTracking) {
					Xm.LOG.info("id=%d addListenger scheduling packet.", id);
				}

				listener.field_13987.method_14364(S2C_ExcavationRenderUpdate.toPacket(this));
			}
		}
	}

	public void removeListener(class_3222 listener) {
		synchronized (listeners) {
			listeners.removeIfPresent(listener);
		}
	}

	public void updateListeners() {
		if (listeners.isEmpty()) {
			return;
		}

		//        // think network operations need to run in world tick
		//        WorldTaskManager.enqueueImmediate(new Runnable()
		//        {
		final class_2596<?> packet = ExcavationRenderEntry.this.isValid && ExcavationRenderEntry.this.positions.size() > 0
			// update
			? S2C_ExcavationRenderUpdate.toPacket(ExcavationRenderEntry.this)
			// remove
			: S2C_ExcavationRenderUpdate.toPacket(ExcavationRenderEntry.this.id);

		//            @Override
		//            public void run()
		//            {
		synchronized (ExcavationRenderEntry.this.listeners) {
			if (!ExcavationRenderEntry.this.listeners.isEmpty()) {
				for (final class_3222 player : listeners) {
					player.field_13987.method_14364(packet);
				}
			}
		}

		//            }
		//        });
	}

	/**
	 * Will be non-null if should render individual renderPositions. Populated when
	 * position count is small enough not to be a problem.
	 */
	@Nullable
	public class_2338[] renderPositions() {
		if (XmConfig.logExcavationRenderTracking) {
			Xm.LOG.info("id %d Render position retrieval, count = %d", id, renderPositions == null ? 0 : renderPositions.length);
		}

		return renderPositions;
	}
}
